* plsemaph.prg
*
* by Joe Gotthelf

para inAction, inSemaph

private all like l_*

l_action = lower(alltrim(inAction))

l_oldrepro = set("repro")
l_olderr = on("error")
l_oldsele = select()

set repro to 0
on error l_x = 1 && Disable existing error handler.
do case
  case l_action = "lock"
    l_RetVal = i_SemLock((inSemaph))
  case l_action = "unlock"
    l_RetVal = i_SemUnl((inSemaph))
  case l_action = "cleanup"
    l_RetVal = i_cleanup()
  otherwise
    l_RetVal = .f.
endcase
if empty(l_olderr) && restore the error handler
  on error
else
  on error &l_olderr.
endif
select (l_oldsele) && restore the selected work area
set repro to l_oldrepro && restore SET REPROCESS value
return l_RetVal

**************************************************
proc i_SemLock
*
* create/lock a semaphore
*
para inSemaph

private all like l_*

if type("g_semdbf") = "U"
  **** if we don't have our "lock dbf", create it.
  public g_semdbf, g_semalia
  g_semdbf = sys(3)
  do while file(g_semdbf)
    g_semdbf = sys(3)
  enddo
  **** following will create & open file exclusive
  create table (g_semdbf) (frog c(1))
  g_semalia = alias()
endif

**** first see if it's in the main semaphore dbf.
if !used("plSemDBF")
  **** open the semaphore table.
  use plSemDBF order sm_name in select(1)
endif
select plSemDBF
set order to sm_name && index is UNIQUE
if seek(padr(inSemaph, len(sm_name)))
  **** The semaphore is already in the .DBF, let's see
  **** if it is actually in use.
  if !rlock()
    **** It is locked by someone else, so we can't
    **** get it.
    return .f.
  endif
  if (sm_dbf == g_semdbf) or ;
     i_canhave(alltrim(sm_dbf))
    **** if it is our own semaphore, or it is not in
    **** use by another active session, it is ours.
    ****
    **** note: I_CANHAVE may now have left the
    ****  (sm_dbf) file open

    **** put our lock .dbf name in, so other's can't
    **** get it.
    replace sm_dbf with g_semdbf
    l_retval = .t.
  else
    **** not our semaphore, or the I_CANHAVE(...)
    **** procedure returned a .F.
    l_retval = .f.
  endif
  unlock && free up the LAN resources
else
  **** Add & mark the semaphore to the semaphore DBF
  insert into plSemDBF values (inSemaph, g_semdbf)
  if !rlock()
    **** some other turkey grabbed it!
    return .f. && everybody fails.
  endif
  l_rec = recno()
  seek padr(inSemaph, len(sm_name))
  if l_rec = recno()
    l_retval = .t.
  else
    **** Record number changed; somebody else added
    **** the same semaphore at the same time.
    goto l_rec
    dele && leave for later packing
    l_retval = .f.
  endif
  unlock && free network resources
endif
if used("OTHSEMAPH")
  **** If we managed to open an old LOCK .DBF, we
  **** will delete it, just to be neat and clean.
  l_othdbf = dbf("OTHSEMAPH")
  use in OTHSEMAPH
  erase (l_othdbf+".dbf")  && clean up
endif
return l_retval

****************************************************
proc i_SemUnl
*
* Unlock a semaphore
*
para inSemaph

if type("g_semdbf") = "U"
  **** we haven't create any semaphores!
  return .f.
endif


if !used("plSemDBF")
  **** open the semaphore table.
  use plSemDBF order sm_name in select(1)
endif
select plSemDBF
set order to sm_name && index is UNIQUE
if seek(padr(inSemaph, len(sm_name)))
  if (sm_dbf == g_semdbf)
    do while !rlock()
      **** nothing will keep it locked very long,
      *** we will be patient....
    enddo
    replace sm_name with ""
    unlock
    return .t.
  else
    **** Not ours to close!
    return .f.
  endif
else
  **** Not even there!
  return .f.
endif

****************************************************
proc i_cleanup
*
* remove the lock .dbf from the system.
*
if used(g_semalia)
  use in (g_semalia)
  erase (g_semdbf+".dbf")
endif
if used("plSemDBF")
   select plSemDBF
   set order to 0
   scan for sm_dbf = g_semdbf
     replace sm_name with ""
     delete && for later packing
   endscan
   use
endif      
release g_semdbf, g_semalia
return .t.

***************************************************
proc i_canhave
*
* determine if the "locking .dbf" associated with a
* semaphor is still in "EXCLUSIVE USE" by its owner.
*
para inDbf

private all like l_*

if !file(inDBF+".dbf")
  **** If the file doesn't exist, the owner gone...
  **** The semaphore is all ours.
  return .t.
else
  **** Try to use it
  use (inDBF) in select(1) alias OTHSEMAPH
  if used("OTHSEMAPH")
    **** We know its owner is no longer using it.
    return .t.
  endif
endif
**** Only happens if the file exists and we can not
**** USE it.
return .f.
